import functools

from django.db import models
import graphene
from graphene_django.compat import (
    ArrayField,
    HStoreField,
    JSONField,
    PGJSONField,
    RangeField,
)
from graphene_django.registry import get_global_registry

_registry = get_global_registry()


class FieldKind(graphene.Enum):
    """Field kind."""

    ID = "id"
    JSON = "json"
    STRING = "string"
    TEXT = "text"
    BOOLEAN = "boolean"
    INTEGER = "integer"
    DECIMAL = "decimal"
    FLOAT = "float"
    DATE = "date"
    DATETIME = "datetime"
    TIME = "time"
    PERCENT = "percent"
    EMAIL = "email"
    SLUG = "slug"
    PHONE = "phone"
    UUID = "uuid"
    IP = "ip"
    URL = "url"
    FILE = "file"
    PASSWORD = "password"
    CURRENCY = "currency"
    POSTAL_CODE = "postal-code"
    COMPANY_DOCUMENT = "company-document"
    INDIVIDUAL_DOCUMENT = "individual-document"


@functools.singledispatch
def get_field_schema(field) -> dict:
    raise Exception(
        "Don't know how to convert the Django field {} ({})".format(field, field.__class__)
    )


@get_field_schema.register(models.CharField)
def get_field_schema_string(field):
    return {
        "kind": FieldKind.STRING,
    }


@get_field_schema.register(models.TextField)
def get_field_schema_text(field):
    return {
        "kind": FieldKind.TEXT,
    }


@get_field_schema.register(models.EmailField)
def get_field_schema_email(field):
    return {
        "kind": FieldKind.EMAIL,
    }


@get_field_schema.register(models.SlugField)
def get_field_schema_slug(field):
    return {
        "kind": FieldKind.SLUG,
    }


@get_field_schema.register(models.UUIDField)
def get_field_schema_uuid(field):
    return {
        "kind": FieldKind.UUID,
    }


@get_field_schema.register(models.URLField)
def get_field_schema_url(field):
    return {
        "kind": FieldKind.URL,
    }


@get_field_schema.register(models.GenericIPAddressField)
def get_field_schema_ip(field):
    return {
        "kind": FieldKind.IP,
    }


@get_field_schema.register(models.FileField)
@get_field_schema.register(models.ImageField)
@get_field_schema.register(models.FilePathField)
def get_field_schema_file(field):
    return {
        "kind": FieldKind.FILE,
    }


@get_field_schema.register(models.AutoField)
@get_field_schema.register(models.PositiveIntegerField)
@get_field_schema.register(models.PositiveSmallIntegerField)
@get_field_schema.register(models.SmallIntegerField)
@get_field_schema.register(models.BigIntegerField)
@get_field_schema.register(models.IntegerField)
def get_field_schema_int(field):
    return {
        "kind": FieldKind.INTEGER,
    }


@get_field_schema.register(models.DecimalField)
@get_field_schema.register(models.DurationField)
def get_field_schema_decimal(field):
    return {
        "kind": FieldKind.DECIMAL,
        "validation": {
            "max_digits": field.max_digits,
            "decimal_places": field.decimal_places,
        },
    }


@get_field_schema.register(models.FloatField)
def get_field_schema_float(field):
    return {
        "kind": FieldKind.FLOAT,
    }


@get_field_schema.register(models.BooleanField)
def get_field_schema_bool(field):
    return {
        "kind": FieldKind.BOOLEAN,
    }


if hasattr(models, "NullBooleanField"):

    @get_field_schema.register(models.NullBooleanField)  # type:ignore
    def get_field_schema_nullbool(field):
        return {
            "kind": FieldKind.BOOLEAN,
        }


@get_field_schema.register(models.DateField)
def get_field_schema_date(field):
    return {
        "kind": FieldKind.DATE,
    }


@get_field_schema.register(models.DateTimeField)
def get_field_schema_datetime(field):
    return {
        "kind": FieldKind.DATETIME,
    }


@get_field_schema.register(models.TimeField)
def get_field_schema_time(field):
    return {
        "kind": FieldKind.TIME,
    }


@get_field_schema.register(models.ForeignKey)
@get_field_schema.register(models.OneToOneRel)
@get_field_schema.register(models.OneToOneField)
def get_field_schema_fk(field):
    model = field.related_model
    _type = _registry.get_type_for_model(model)
    return {
        "kind": FieldKind.ID,
        "of_type": _type._meta.name if _type else None,
    }


@get_field_schema.register(models.ManyToManyField)
@get_field_schema.register(models.ManyToManyRel)
@get_field_schema.register(models.ManyToOneRel)
def get_field_schema_m2m(field):
    model = field.related_model
    _type = _registry.get_type_for_model(model)
    return {
        "kind": FieldKind.ID,
        "of_type": _type._meta.name if _type else None,
        "multiple": True,
    }


@get_field_schema.register(ArrayField)
@get_field_schema.register(RangeField)
def get_field_schema_array(field):
    d = get_field_schema(field.base_field)
    d["multiple"] = True
    return d


@get_field_schema.register(PGJSONField)
@get_field_schema.register(JSONField)
@get_field_schema.register(HStoreField)
def get_field_schema_pg(field):
    return {
        "kind": FieldKind.JSON,
    }



"""
Settings for graphene-django-plus are all namespaced in the GRAPHENE_DJANGO_PLUS setting.
For example your project's `settings.py` file might look like this:
GRAPHENE_DJANGO_PLUS = {
    'MUTATIONS_INCLUDE_REVERSE_RELATIONS': False
}
This module provides the `graphene_django_plus_settings` object, that is used to access
graphene-django-plus settings, checking for user settings first, then falling
back to the defaults.
"""
import importlib

from django.conf import settings
from django.test.signals import setting_changed

# Copied shamelessly from Django REST Framework and graphene-django

DEFAULTS = {
    "MUTATIONS_INCLUDE_REVERSE_RELATIONS": True,
    "MUTATIONS_SWALLOW_PERMISSION_DENIED": True,
}

# List of settings that may be in string import notation.
IMPORT_STRINGS = []


def perform_import(val, setting_name):
    """
    If the given setting is a string import notation,
    then perform the necessary import or imports.
    """
    if val is None:
        return None
    elif isinstance(val, str):
        return import_from_string(val, setting_name)
    elif isinstance(val, (list, tuple)):
        return [import_from_string(item, setting_name) for item in val]
    return val


def import_from_string(val, setting_name):
    """
    Attempt to import a class from a string representation.
    """
    try:
        # Nod to tastypie's use of importlib.
        parts = val.split(".")
        module_path, class_name = ".".join(parts[:-1]), parts[-1]
        module = importlib.import_module(module_path)
        return getattr(module, class_name)
    except (ImportError, AttributeError) as e:
        msg = "Could not import '{}' for graphene-django-plus setting '{}'. {}: {}.".format(
            val,
            setting_name,
            e.__class__.__name__,
            e,
        )
        raise ImportError(msg)


class GrapheneDjangoPlusSettings:
    """
    A settings object, that allows API settings to be accessed as properties.
    For example:
        from graphene_django_plus.settings import settings
        print(settings.MUTATIONS_INCLUDE_REVERSE_RELATIONS)
    Any setting with string import paths will be automatically resolved
    and return the class, rather than the string literal.
    """

    def __init__(self, user_settings=None, defaults=None, import_strings=None):
        if user_settings:
            self._user_settings = user_settings
        self.defaults = defaults or DEFAULTS
        self.import_strings = import_strings or IMPORT_STRINGS
        self._cached_attrs = set()

    @property
    def user_settings(self):
        if not hasattr(self, "_user_settings"):
            self._user_settings = getattr(settings, "GRAPHENE_DJANGO_PLUS", {})
        return self._user_settings

    def reload(self):
        for attr in self._cached_attrs:
            delattr(self, attr)
        self._cached_attrs.clear()
        if hasattr(self, "_user_settings"):
            delattr(self, "_user_settings")

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError("Invalid graphene-django-plus setting: '%s'" % attr)

        try:
            # Check if present in user settings
            val = self.user_settings[attr]
        except KeyError:
            # Fall back to defaults
            val = self.defaults[attr]

        # Coerce import strings into classes
        if attr in self.import_strings:
            val = perform_import(val, attr)

        # Cache the result
        self._cached_attrs.add(attr)
        setattr(self, attr, val)
        return val


graphene_django_plus_settings = GrapheneDjangoPlusSettings(None, DEFAULTS, IMPORT_STRINGS)


def reload_graphene_django_plus_settings(*args, **kwargs):
    setting = kwargs["setting"]
    if setting == "GRAPHENE_DJANGO_PLUS":
        graphene_django_plus_settings.reload()


setting_changed.connect(reload_graphene_django_plus_settings)



try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable
from typing import TYPE_CHECKING, List, Tuple, Type, TypeVar, Union

try:
    from guardian.conf import settings as guardian_settings
    from guardian.core import ObjectPermissionChecker
    from guardian.shortcuts import get_objects_for_user

    has_guardian = True
except ImportError:  # pragma: no cover
    has_guardian = False

from django.apps import apps
from django.contrib.auth.models import AbstractUser, AnonymousUser
from django.db import models
from django.db.models.query import QuerySet

if TYPE_CHECKING:  # pragma: nocover
    from guardian.core import ObjectPermissionChecker  # noqa: F811

_T = TypeVar("_T", bound="GuardedModel")
_TR = TypeVar("_TR", bound="GuardedRelatedModel")


def _separate_perms(
    perms: Iterable,
    model: Type[models.Model],
) -> Tuple[List[str], List[str]]:
    perms = set(perms)
    own_perms = perms & (
        {p[0] for p in model._meta.permissions}
        | {f"{model._meta.app_label}.{p[0]}" for p in model._meta.permissions}
    )
    other_perms = perms - own_perms
    return list(other_perms), list(own_perms)


def _has_anonymous_user():
    if not has_guardian:
        return False

    return guardian_settings.ANONYMOUS_USER_NAME is not None


class GuardedModelManager(models.Manager[_T]):
    """Model manager that integrates with guardian to check for permissions."""

    def for_user(
        self,
        user: Union[AbstractUser, AnonymousUser],
        perms: Union[str, List[str]],
        any_perm: bool = True,
        with_superuser: bool = True,
    ) -> QuerySet[_T]:
        """Get a queryset filtered by perms for the user.

        :param user: the user itself
        :param perms: a string or list of perms to check for
        :param any_perm: if any perm or all perms should be considered
        :param with_superuser: if a superuser should skip the checks
        """
        # No guardian means we are not checking perms
        if not has_guardian:
            return self.all()

        # In case the user is anonymous we don't have an anonymous user set,
        # then we can't return anything.
        if isinstance(user, AnonymousUser) and not _has_anonymous_user():
            return self.none()

        perms = [perms] if isinstance(perms, str) else perms
        perms = [p.split(".", 1)[1] if "." in p else p for p in perms]

        return get_objects_for_user(
            user,
            perms,
            klass=self.model,
            any_perm=any_perm,
            with_superuser=with_superuser,
        )


class GuardedRelatedManager(GuardedModelManager[_TR]):
    """Manager for objects related to companies."""

    model: Type[_TR]

    def for_user(
        self,
        user: Union[AbstractUser, AnonymousUser],
        perms: Union[str, List[str]],
        any_perm: bool = True,
        with_superuser: bool = True,
    ) -> models.QuerySet[_TR]:
        # No guardian means we are not checking perms
        if not has_guardian:
            return self.all()

        # In case the user is anonymous we don't have an anonymous user set,
        # then we can't return anything.
        if isinstance(user, AnonymousUser) and not _has_anonymous_user():
            return self.none()

        perms = [perms] if isinstance(perms, str) else perms
        other_perms, own_perms = _separate_perms(perms, self.model)

        if own_perms:
            own_qs = super().for_user(
                user,
                own_perms,
                any_perm=any_perm,
                with_superuser=with_superuser,
            )
        else:
            own_qs = None

        if other_perms:
            m = self.model.related_model
            if isinstance(m, str):
                app_label, model_name = m.split(".")
                m = apps.get_model(app_label=app_label, model_name=model_name)

            other_qs = self.all().filter(
                **{
                    f"{self.model.related_attr}__in": m.objects.for_user(
                        user,
                        other_perms,
                        any_perm=any_perm,
                        with_superuser=with_superuser,
                    ),
                }
            )
        else:
            other_qs = None

        if own_qs is not None and other_qs is not None:
            if any_perm:
                return own_qs | other_qs
            else:
                return own_qs & other_qs
        elif own_qs is not None:
            return own_qs
        elif other_qs is not None:
            return other_qs
        else:  # pragma: nocover
            raise AssertionError


class GuardedModel(models.Model):
    """Model that integrates with guardian to check for permissions."""

    class Meta:
        abstract = True

    # NOTE: This should be adjusted in each model using this class
    # for better typing support
    objects = GuardedModelManager["GuardedModel"]()

    def has_perm(
        self,
        user: Union[AbstractUser, AnonymousUser],
        perms: Union[str, List[str]],
        any_perm: bool = True,
        checker: Union[ObjectPermissionChecker, None] = None,
    ) -> bool:
        """Check if the user has the given permissions to this object.

        :param user: the user itself
        :param perms: a string or list of perms to check for
        :param any_perm: if any perm or all perms should be considered
        :param checker: a `guardian.core.ObjectPermissionChecker` that
            can be used to optimize performance in checking for permissions
        """
        # No guardian means we are not checking perms
        if not has_guardian:
            return True

        # In case the user is anonymous we don't have an anonymous user set,
        # then we can't return anything.
        if isinstance(user, AnonymousUser) and not _has_anonymous_user():
            return False

        perms = [perms] if isinstance(perms, str) else perms
        checker = checker or ObjectPermissionChecker(user)
        f = any if any_perm else all

        # First try to check if the user has global permissions for this
        # Otherwise we will check for the object itself bellow
        if f(user.has_perm(p) for p in perms):
            return True

        # Small performance improvement by mimicing guardian's api
        perms = [p.split(".", 1)[1] if "." in p else p for p in perms]
        c_perms = checker.get_perms(self)
        return f(p in c_perms for p in perms)


class GuardedRelatedModel(GuardedModel):
    """Base model for objects that are related to other guarded model."""

    class Meta:
        abstract = True

    objects = GuardedRelatedManager["GuardedRelatedModel"]()
    related_model: Union[Type[GuardedModel], str]
    related_attr: str

    def has_perm(
        self,
        user: Union[AbstractUser, AnonymousUser],
        perms: Union[str, List[str]],
        any_perm: bool = True,
        checker: Union[ObjectPermissionChecker, None] = None,
    ) -> bool:
        # No guardian means we are not checking perms
        if not has_guardian:
            return True

        # In case the user is anonymous we don't have an anonymous user set,
        # then we can't return anything.
        if isinstance(user, AnonymousUser) and not _has_anonymous_user():
            return False

        perms = [perms] if isinstance(perms, str) else perms
        checker = checker or ObjectPermissionChecker(user)
        other_perms, own_perms = _separate_perms(perms, self.__class__)

        if own_perms:
            super_has_perm = super().has_perm
            own_check = lambda: super_has_perm(
                user,
                list(own_perms),
                any_perm=any_perm,
                checker=checker,
            )
        else:
            own_check = lambda: not any_perm

        if other_perms:
            assert self.related_attr is not None
            related = getattr(self, self.related_attr)
            other_check = lambda: related.has_perm(
                user,
                other_perms,
                any_perm=any_perm,
                checker=checker,
            )
        else:
            other_check = lambda: not any_perm

        f = any if any_perm else all
        # Using lambdas will shortcut own_check when own_check is True
        # and we are checking for any_perm
        return f(check() for check in [other_check, own_check])



import contextlib
import json

from graphene_django.views import GraphQLView as _GraphQLView


def _get_key(key):
    try:
        int_key = int(key)
    except (TypeError, ValueError):
        return key
    else:
        return int_key


def _get_shallow_property(obj, prop):
    if isinstance(prop, int):
        return obj[prop]

    try:
        return obj.get(prop)
    except AttributeError:
        return None


def _obj_set(obj, path, value):
    if isinstance(path, int):
        path = [path]

    if not path:
        return obj

    if isinstance(path, str):
        new_path = [_get_key(part) for part in path.split(".")]
        return _obj_set(obj, new_path, value)

    current_path = path[0]
    current_value = _get_shallow_property(obj, current_path)

    if len(path) == 1:
        obj[current_path] = value

    if current_value is None:
        with contextlib.suppress(IndexError):
            if isinstance(path[1], int):
                obj[current_path] = []
            else:
                obj[current_path] = {}

    return _obj_set(obj[current_path], path[1:], value)


class GraphQLView(_GraphQLView):
    """GraphQLView with file upload support.

    Based on:
        https://github.com/mirumee/saleor/blob/master/saleor/graphql/views.py
    """

    @staticmethod
    def get_graphql_params(request, data):
        query, variables, operation_name, id_ = _GraphQLView.get_graphql_params(
            request,
            data,
        )

        content_type = _GraphQLView.get_content_type(request)
        if content_type == "multipart/form-data":
            operations = json.loads(data.get("operations", "{}"))
            files_map = json.loads(data.get("map", "{}"))
            for k, v in files_map.items():
                for f in v:
                    _obj_set(operations, f, k)
            query = operations.get("query")
            variables = operations.get("variables")

        return query, variables, operation_name, id_



import graphene

from .types import SchemaType, schema_registry


class Query:
    """Queries object."""

    gql_object_schema = graphene.Field(
        SchemaType,
        description="GraphQL input schema for forms.",
        object_type=graphene.String(
            description="The input object to query for.",
            required=True,
        ),
        required=False,
        default_value=None,
    )
    gql_object_schema_all = graphene.List(
        graphene.NonNull(SchemaType),
        description="GraphQL input schema for forms.",
        required=True,
    )

    @staticmethod
    def resolve_gql_object_schema(root, info, object_type, **kwargs):
        return schema_registry.get(object_type, None)

    @staticmethod
    def resolve_gql_object_schema_all(root, info, **kwargs):
        return sorted(schema_registry.values(), key=lambda obj: obj["object_type"])



import datetime
import decimal
from typing import TYPE_CHECKING, Generic, List, Optional, TypeVar, Union

from django.contrib.auth.models import AbstractUser, AnonymousUser
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.db.models import Prefetch
from django.db.models.fields import NOT_PROVIDED
from django.db.models.fields.reverse_related import ManyToManyRel, ManyToOneRel
import graphene
from graphene.utils.str_converters import to_camel_case
from graphene_django import DjangoObjectType
from graphene_django.converter import get_choices
from graphene_django.types import DjangoObjectTypeOptions

try:
    import graphene_django_optimizer as gql_optimizer

    _BaseDjangoObjectType = gql_optimizer.OptimizedDjangoObjectType
except ImportError:
    gql_optimizer = None
    _BaseDjangoObjectType = DjangoObjectType

from .models import GuardedModel, GuardedModelManager
from .perms import check_authenticated, check_perms
from .schema import FieldKind, get_field_schema
from .utils import get_model_fields, update_dict_nested

if TYPE_CHECKING:  # pragma: nocover
    _BaseDjangoObjectType = DjangoObjectType

_T = TypeVar("_T", bound=models.Model)
schema_registry = {}


def schema_for_field(field, name):
    s = get_field_schema(field)

    default_value = getattr(field, "default", None)
    if default_value is NOT_PROVIDED:
        default_value = None
    if default_value is not None and callable(default_value):
        default_value = default_value()
    if default_value is not None:
        if isinstance(default_value, decimal.Decimal):
            default_value = str(default_value)
        if isinstance(default_value, (datetime.datetime, datetime.date, datetime.time)):
            default_value = default_value.isoformat()

    if isinstance(field, (ManyToOneRel, ManyToManyRel)):
        required = not field.null
    else:
        required = not field.blank and field.default is NOT_PROVIDED

    items = getattr(field, "choices", None)
    if items:
        if isinstance(items, dict):  # pragma:nocover
            items = items.items()

        choices = []
        for (_original_v, label), (n, _value, _desc) in zip(items, get_choices(items)):
            choices.append(
                {
                    "label": label,
                    "value": n,
                }
            )
    else:
        choices = None

    s = update_dict_nested(
        s,
        {
            "name": to_camel_case(name),
            # FIXME: Get verbose_name and help_text for m2m
            "label": getattr(field, "verbose_name", None),
            "help_text": getattr(field, "help_text", None),
            "hidden": name == "id",
            "choices": choices,
            "default_value": default_value,
            "validation": {
                "required": required,
                "min_length": getattr(field, "min_length", None),
                "max_length": getattr(field, "max_length", None),
                "min_value": None,
                "max_value": None,
            },
        },
    )

    return s


class MutationErrorType(graphene.ObjectType):
    """An error that happened in a mutation."""

    field = graphene.String(
        description=(
            "The field that caused the error, or `null` if it "
            "isn't associated with any particular field."
        ),
        required=False,
    )
    message = graphene.String(
        description="The error message.",
    )


class InputSchemaFieldChoiceType(graphene.ObjectType):
    """An input schema field choice."""

    label = graphene.String(
        description="The choice's label.",
        required=True,
    )
    value = graphene.String(
        description="The choice's value.",
        required=True,
    )


class SchemaFieldValidationType(graphene.ObjectType):
    """Validation data for the field."""

    required = graphene.Boolean(
        description="If this field is required.",
        required=True,
        default_value=False,
    )
    min_value = graphene.JSONString(
        description="Min value for the field. Parse the json to get its value.",
        required=False,
        default_value=None,
    )
    max_value = graphene.JSONString(
        description="Max value for the field. Parse the json to get its value.",
        required=False,
        default_value=None,
    )
    min_length = graphene.Int(
        description="Min length for string kinds.",
        required=False,
        default_value=None,
    )
    max_length = graphene.Int(
        description="Max length for string kinds.",
        required=False,
        default_value=None,
    )
    max_digits = graphene.Int(
        description="Max digits for decimal kinds (null otherwise).",
        required=False,
        default_value=None,
    )
    decimal_places = graphene.Int(
        description="Max digits for decimal kinds (null otherwise).",
        required=False,
        default_value=None,
    )


class SchemaFieldType(graphene.ObjectType):
    """The input schema field."""

    name = graphene.String(
        description="The name of the field",
        required=True,
    )
    kind = FieldKind(
        description="The kind of this field.",
        required=True,
    )
    of_type = graphene.String(
        description="The name of the related field for ID kinds.",
        required=False,
        default_value=None,
    )
    multiple = graphene.Boolean(
        description="If this field expects an array of values.",
        required=True,
        default_value=False,
    )
    choices = graphene.List(
        graphene.NonNull(InputSchemaFieldChoiceType),
        description="Choices for this field.",
        required=False,
        default_value=None,
    )
    hidden = graphene.Boolean(
        description="If this field should be displayed in a hidden field.",
        required=True,
        default_value=False,
    )
    label = graphene.String(
        description="The field's humanized name.",
        required=False,
        default_value=None,
    )
    help_text = graphene.String(
        description="A help text for the field.",
        required=False,
        default_value=None,
    )
    default_value = graphene.JSONString(
        description="Default value for the field. Parse the json to get its value.",
        required=False,
        default_value=None,
    )
    validation = graphene.Field(
        SchemaFieldValidationType,
        description="Validation metadata for this field.",
        required=True,
    )


class SchemaType(graphene.ObjectType):
    """The input schema."""

    object_type = graphene.String(
        description="The name of the input object.",
        required=True,
    )
    fields = graphene.List(
        graphene.NonNull(SchemaFieldType),
        description="The fields in the input object.",
        required=True,
    )


class UploadType(graphene.types.Scalar):
    """The upload of a file.

    Variables of this type must be set to null in mutations. They will be
    replaced with a filename from a following multipart part containing a
    binary file.

    See: https://github.com/jaydenseric/graphql-multipart-request-spec
    """

    @staticmethod
    def serialize(value):
        return value

    @staticmethod
    def parse_literal(node):
        return node

    @staticmethod
    def parse_value(value):
        return value


class ModelTypeOptions(DjangoObjectTypeOptions):
    """Model type options for :class:`ModelType`."""

    #: If we shuld allow unauthenticated users to query for this model.
    public: bool = False

    #: A list of django permissions to check if the user has permission to
    #: query this model.
    permissions: Optional[List[str]] = None

    #: If any permission should allow the user to query this model.
    permissions_any: bool = True

    #: A list of guardian object permissions to check if the user has
    #: permission to query the model object.
    object_permissions: Optional[List[str]] = None

    #: If any object permission should allow the user to query this model.
    object_permissions_any: bool = True

    #: If superuser should be considered when getting `GuardedModelManager.for_user`
    object_permissions_with_superuser: bool = True

    #: The fields schema for the schema query
    fields_schema: Optional[dict] = None


class ModelType(_BaseDjangoObjectType, Generic[_T]):
    """Base type with automatic optimizations and permissions checking."""

    class Meta:
        abstract = True

    @classmethod
    def __class_getitem__(cls, *args, **kwargs):
        return cls

    @classmethod
    def __init_subclass_with_meta__(
        cls,
        _meta=None,
        model=None,
        permissions=None,
        permissions_any=True,
        object_permissions=None,
        object_permissions_any=True,
        object_permissions_with_superuser=True,
        fields_schema=None,
        public=None,
        only_fields=None,
        fields=None,
        exclude_fields=None,
        exclude=None,
        **kwargs,
    ):
        if not _meta:
            _meta = ModelTypeOptions(cls)

        if "allow_unauthenticated" in kwargs:
            raise ImproperlyConfigured("Use 'public' instead of 'allow_unauthenticated'")

        _meta.permissions = permissions or []
        _meta.permissions_any = permissions_any
        _meta.object_permissions = object_permissions or []
        _meta.object_permissions_any = object_permissions_any
        _meta.object_permissions_with_superuser = object_permissions_with_superuser
        _meta.public = public

        _fields_schema = {}
        # graphene will handle the deprecated only_fields/exclude_fields for us
        # We just want to mimic the logic here
        _include = fields if fields is not None else only_fields
        if _include is not None:
            _include = set(_include)
        _exclude = set(exclude or []) or set(exclude_fields or [])
        for name, field in get_model_fields(model):
            if name in _exclude:
                continue
            if _include is not None and name not in _include:
                continue

            _fields_schema[name] = schema_for_field(field, name)

        fields_schema = update_dict_nested(
            _fields_schema,
            fields_schema or {},
        )
        _meta.fields_schema = fields_schema or {}

        super().__init_subclass_with_meta__(
            _meta=_meta,
            model=model,
            only_fields=only_fields,
            fields=fields,
            exclude_fields=exclude_fields,
            exclude=exclude,
            **kwargs,
        )

        schema_registry[cls._meta.name] = {
            "object_type": cls._meta.name,
            "fields": list(_meta.fields_schema.values()),
        }

    @classmethod
    def get_queryset(
        cls,
        qs: Union[models.QuerySet[_T], models.Manager[_T]],
        info,
    ) -> models.QuerySet[_T]:
        """Get the queryset checking for permissions and optimizing the query.

        Override the default graphene's `get_queryset` to check for permissions
        and optimize the query performance.

        Note that the query will only be automaticallu optimized if,
        `graphene_django_optimizer` is installed.
        """
        if isinstance(qs, models.Manager):
            qs = qs.get_queryset()

        if not cls.check_permissions(info.context.user):
            return qs.none()

        if cls._meta.object_permissions and isinstance(
            cls._meta.model.objects, GuardedModelManager
        ):
            qs &= cls._meta.model.objects.for_user(
                info.context.user,
                cls._meta.object_permissions,
                any_perm=cls._meta.object_permissions_any,
                with_superuser=cls._meta.object_permissions_with_superuser,
            )

        ret = super().get_queryset(qs, info)
        if gql_optimizer is not None:
            ret = gql_optimizer.query(ret, info)
            prl = {
                i.to_attr if isinstance(i, Prefetch) else i: i  # type:ignore
                for i in ret._prefetch_related_lookups
            }
            ret._prefetch_related_lookups = tuple(prl.values())

        return ret

    @classmethod
    def get_node(cls, info, id_) -> Optional[_T]:
        """Get the node instance given the relay global id."""
        # NOTE: get_queryset will filter allowed models for the user so
        # this will return None if he is not allowed to retrieve this

        if gql_optimizer is not None:
            # optimizer will ignore queryset so call the same as they call
            # but passing objects from get_queryset to keep our preferences
            try:
                instance = cls.get_queryset(cls._meta.model.objects, info).get(
                    pk=id_,
                )
            except cls._meta.model.DoesNotExist:
                instance = None
        else:
            instance = super().get_node(info, id_)

        if instance is not None and not cls.check_object_permissions(info.context.user, instance):
            return None

        return instance

    @classmethod
    def check_permissions(cls, user: Union[AbstractUser, AnonymousUser]) -> bool:
        """Check permissions for the given user.

        Subclasses can override this to avoid the permission checking or
        extending it. Remember to call `super()` in the later case.
        """
        if not cls._meta.public and not check_authenticated(user):
            return False

        if not cls._meta.permissions:
            return True

        return check_perms(user, cls._meta.permissions, any_perm=cls._meta.permissions_any)

    @classmethod
    def check_object_permissions(
        cls,
        user: Union[AbstractUser, AnonymousUser],
        instance: _T,
    ) -> bool:
        """Check object permissions for the given user.

        Subclasses can override this to avoid the permission checking or
        extending it. Remember to call `super()` in the later case.

        For this to work, the model needs to implement a `has_perm` method.
        The easiest way when using `guardian` is to inherit it
        from :class:`graphene_django_plus.models.GuardedModel`.
        """
        if not cls._meta.object_permissions:
            return True

        if not isinstance(instance, GuardedModel):
            return True

        return instance.has_perm(
            user,
            cls._meta.object_permissions,
            any_perm=cls._meta.object_permissions_any,
        )



from typing import List, Optional, Union

from django.contrib.auth.models import AbstractUser, AnonymousUser

from .exceptions import PermissionDenied


def check_authenticated(user: Union[AbstractUser, AnonymousUser]):
    return user and user.is_authenticated


def assert_authenticated(user: Union[AbstractUser, AnonymousUser], msg: Optional[str] = None):
    if not check_authenticated(user):
        raise PermissionDenied(msg or "You don't have permissions to do this...")


def check_superuser(user: Union[AbstractUser, AnonymousUser]):
    return check_authenticated(user) and user.is_superuser


def assert_superuser(user: Union[AbstractUser, AnonymousUser], msg: Optional[str] = None):
    if not check_superuser(user):
        raise PermissionDenied(msg or "You don't have permissions to do this...")


def check_perms(
    user: Union[AbstractUser, AnonymousUser],
    perms: List[str],
    any_perm: bool = True,
    with_superuser: bool = True,
):
    if not check_authenticated(user):
        return False

    if with_superuser and check_superuser(user):
        return True

    u_perms = set(user.get_all_permissions())
    f = any if any_perm else all
    return f(p in u_perms for p in perms)


def assert_perms(
    user: Union[AbstractUser, AnonymousUser],
    perms: List[str],
    any_perm: bool = True,
    with_superuser: bool = True,
    msg: Optional[str] = None,
):
    if not check_perms(user, perms, any_perm=any_perm, with_superuser=with_superuser):
        raise PermissionDenied(msg or "You don't have permissions to do this...")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['graphene_django_plus']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'graphene-django-plus',
    'version': '3.1.3',
    'description': 'Tools to easily create permissioned CRUD endpoints in graphene.',
    'long_description': '# graphene-django-plus #\n\n[![build status](https://img.shields.io/endpoint.svg?url=https%3A%2F%2Factions-badge.atrox.dev%2F0soft%2Fgraphene-django-plus%2Fbadge%3Fref%3Dmaster&style=flat)](https://actions-badge.atrox.dev/0soft/graphene-django-plus/goto?ref=master)\n[![docs status](https://img.shields.io/readthedocs/graphene-django-plus.svg)](https://graphene-django-plus.readthedocs.io)\n[![coverage](https://img.shields.io/codecov/c/github/0soft/graphene-django-plus.svg)](https://codecov.io/gh/0soft/graphene-django-plus)\n[![PyPI version](https://img.shields.io/pypi/v/graphene-django-plus.svg)](https://pypi.org/project/graphene-django-plus/)\n![python version](https://img.shields.io/pypi/pyversions/graphene-django-plus.svg)\n![django version](https://img.shields.io/pypi/djversions/graphene-django-plus.svg)\n\nTools to easily create permissioned CRUD endpoints in [graphene-django](https://github.com/graphql-python/graphene-django).\n\n## Install\n\n```bash\npip install graphene-django-plus\n```\n\nTo make use of everything this lib has to offer, it is recommended to install\nboth [graphene-django-optimizer](https://github.com/tfoxy/graphene-django-optimizer)\nand [django-guardian](https://github.com/django-guardian/django-guardian).\n\n```bash\npip install graphene-django-optimizer django-guardian\n```\n\n## What it does\n\n* Provides some base types for Django Models to improve querying them with:\n    * Unauthenticated user handling\n    * Automatic optimization using [graphene-django-optimizer](https://github.com/tfoxy/graphene-django-optimizer)\n    * Permission handling for queries using the default [django permission system](https://docs.djangoproject.com/en/2.2/topics/auth/default/#topic-authorization)\n    * Object permission handling for queries using [django guardian](https://github.com/django-guardian/django-guardian)\n    * Relay id conversion so querying can use the global id instead of the model\'s id\n* Provides a set of complete and simple CRUD mutations with:\n    * Unauthenticated user handling\n    * Permission handling using the default [django permission system](https://docs.djangoproject.com/en/2.2/topics/auth/default/#topic-authorization)\n    * Object permission handling using [django guardian](https://github.com/django-guardian/django-guardian)\n    * Automatic input generation based on the model (no need to write your own input type or use `django forms` and `drf serializers`)\n    * Automatic model validation based on the model\'s validators\n* Very simple to create some quick CRUD endpoints for your models\n* Easy to extend and override functionalities\n* File upload handling\n\n## What is included\n\nCheck the [docs](https://graphene-django-plus.readthedocs.io) for a complete\napi documentation.\n\n### Models\n\n* `graphene_django_plus.models.GuardedModel`: A django model that can be used\n  either directly or as a mixin. It will provide a `.has_perm` method and a\n  `.objects.for_user` that will be used by `ModelType` described bellow to\n  check for object permissions.  some utilities to check.\n\n### Types and Queries\n\n* `graphene_django_plus.types.ModelType`: This enchances\n  `graphene_django_plus.DjangoModelType` by doing some automatic `prefetch`\n  optimization on setup and also checking for objects permissions on queries\n  when it inherits from `GuardedModel`.\n\n* `graphene_django_plus.fields.CountableConnection`: This enchances\n  `graphene.relay.Connection` to provide a `total_count` attribute.\n\nHere is an example describing how to use those:\n\n```py\nimport graphene\nfrom graphene import relay\nfrom graphene_django.fields import DjangoConnectionField\n\nfrom graphene_django_plus.models import GuardedModel\nfrom graphene_django_plus.types import ModelType\nfrom graphene_django_plus.fields import CountableConnection\n\n\nclass MyModel(GuardedModel):\n    class Meta:\n        # guardian permissions for this model\n        permissions = [\n            (\'can_read\', "Can read the this object\'s info."),\n        ]\n\n    name = models.CharField(max_length=255)\n\n\nclass MyModelType(ModelType):\n    class Meta:\n        model = MyModel\n        interfaces = [relay.Node]\n\n        # Use our CountableConnection\n        connection_class = CountableConnection\n\n        # When adding this to a query, only objects with a `can_read`\n        # permission to the request\'s user will be allowed to return to him\n        # Note that `can_read` was defined in the model.\n        # If the model doesn\'t inherid from `GuardedModel`, `guardian` is not\n        # installed or this list is empty, any object will be allowed.\n        # This is empty by default\n        object_permissions = [\n            \'can_read\',\n        ]\n\n        # If unauthenticated users should be allowed to retrieve any object\n        # of this type. This is not dependant on `GuardedModel` and neither\n        # `guardian` and is defined as `False` by default\n        public = False\n\n        # A list of Django model permissions to check. Different from\n        # object_permissions, this uses the basic Django\'s permission system\n        # and thus is not dependant on `GuardedModel` and neither `guardian`.\n        # This is an empty list by default.\n        permissions = []\n\n\nclass Query(graphene.ObjectType):\n    my_models = DjangoConnectionField(MyModelType)\n    my_model = relay.Node.Field(MyModelType)\n```\n\nThis can be queried like:\n\n```graphql\n# All objects that the user has permission to see\nquery {\n  myModels {\n    totalCount\n    edges {\n      node {\n        id\n        name\n      }\n    }\n  }\n}\n\n# Single object if the user has permission to see it\nquery {\n  myModel(id: "<relay global ID>") {\n    id\n    name\n  }\n}\n```\n\n### Mutations\n\n* `graphene_django_plus.mutations.BaseMutation`: Base mutation using `relay`\n  and some basic permission checking. Just override its `.perform_mutation` to\n  perform the mutation.\n\n* `graphene_django_plus.mutations.ModelMutation`: Model mutation capable of\n  both creating and updating a model based on the existence of an `id`\n  attribute in the input. All the model\'s fields will be automatically read\n  from Django, inserted in the input type and validated.\n\n* `graphene_django_plus.mutations.ModelCreateMutation`: A `ModelMutation`\n  enforcing a "create only" rule by excluding the `id` field from the input.\n\n* `graphene_django_plus.mutations.ModelUpdateMutation`: A `ModelMutation`\n  enforcing a "update only" rule by making the `id` field required in the\n  input.\n\n* `graphene_django_plus.mutations.ModelDeleteMutation`: A mutation that will\n  receive only the model\'s id and will delete it (if given permission, of\n  course).\n\nHere is an example describing how to use those:\n\n```py\nimport graphene\nfrom graphene import relay\n\nfrom graphene_django_plus.models import GuardedModel\nfrom graphene_django_plus.types import ModelType\nfrom graphene_django_plus.mutations import (\n    ModelCreateMutation,\n    ModelUpdateMutation,\n    ModelDeleteMutation,\n)\n\n\nclass MyModel(GuardedModel):\n    class Meta:\n        # guardian permissions for this model\n        permissions = [\n            (\'can_write\', "Can update this object\'s info."),\n        ]\n\n    name = models.CharField(max_length=255)\n\n\nclass MyModelType(ModelType):\n    class Meta:\n        model = MyModel\n        interfaces = [relay.Node]\n\n\nclass MyModelUpdateMutation(ModelUpdateMutation):\n    class Meta:\n        model = MyModel\n\n        # Make sure only users with the given permissions can modify the\n        # object.\n        # If the model doesn\'t inherid from `GuardedModel`, `guardian` is not\n        # installed ot this list is empty, any object will be allowed.\n        # This is empty by default.\n        object_permissions = [\n            \'can_write\',\n        ]\n\n        # If unauthenticated users should be allowed to retrieve any object\n        # of this type. This is not dependant on `GuardedModel` and neither\n        # `guardian` and is defined as `False` by default\n        public = False\n\n        # A list of Django model permissions to check. Different from\n        # object_permissions, this uses the basic Django\'s permission system\n        # and thus is not dependant on `GuardedModel` and neither `guardian`.\n        # This is an empty list by default.\n        permissions = []\n\n\nclass MyModelDeleteMutation(ModelDeleteMutation):\n    class Meta:\n        model = MyModel\n        object_permissions = [\n            \'can_write\',\n        ]\n\n\nclass MyModelCreateMutation(ModelCreateMutation):\n    class Meta:\n        model = MyModel\n\n    @classmethod\n    def after_save(cls, info, instance, cleaned_input=None):\n        # If the user created the object, allow him to modify it\n        assign_perm(\'can_write\', info.context.user, instance)\n\n\nclass Mutation(graphene.ObjectType):\n    my_model_create = MyModelCreateMutation.Field()\n    my_model_update = MyModelUpdateMutation.Field()\n    my_model_delete = MyModelDeleteMutation.Field()\n```\n\nThis can be used to create/update/delete like:\n\n```graphql\n# Create mutation\nmutation {\n  myModelCreate(input: {name: "foobar"}) {\n    myModel {\n      name\n    }\n    errors {\n      field\n      message\n    }\n  }\n}\n\n# Update mutation\nmutation {\n  myModelUpdate(input: {id: "<relay global ID>" name: "foobar"}) {\n    myModel {\n      name\n    }\n    errors {\n      field\n      message\n    }\n  }\n}\n\n# Delete mutation\nmutation {\n  myModelDelete(input: {id: "<relay global ID>"}) {\n    myModel {\n      name\n    }\n    errors {\n      field\n      message\n    }\n  }\n}\n```\n\nAny validation errors will be presented in the `errors` return value.\n\nTo turn off auto related relations addition to the mutation input - set global\n`MUTATIONS_INCLUDE_REVERSE_RELATIONS` parameter to `False` in your\n`settings.py`:\n\n```\nGRAPHENE_DJANGO_PLUS = {\n    \'MUTATIONS_INCLUDE_REVERSE_RELATIONS\': False\n}\n```\n\nNote: in case reverse relation does not have `related_name` attribute set -\nmutation input will be generated as Django itself is generating by appending\n`_set` to the lower cased model name - `modelname_set`\n\n## License\n\nThis project is licensed under MIT licence (see `LICENSE` for more info)\n\n## Contributing\n\nMake sure to have [poetry](https://python-poetry.org/) installed.\n\nInstall dependencies with:\n\n```bash\npoetry install\n```\n\nRun the testsuite with:\n\n```bash\npoetry run pytest\n```\n\nFeel free to fork the project and send me pull requests with new features,\ncorrections and translations. We\'ll gladly merge them and release new versions\nASAP.\n',
    'author': 'Thiago Bellini Ribeiro',
    'author_email': 'thiago@bellini.dev',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/0soft/graphene-django-plus',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
